# -*- coding: utf-8 -*-
"""Test tables"""

import unittest

from utils import PyrseasTestCase, fix_indent

TYPELIST = [
    ('SMALLINT', 'smallint'),
    ('INTEGER', 'integer'),
    ('BIGINT', 'bigint'),
    ('int', 'integer'),
    ('int2', 'smallint'),
    ('int4', 'integer'),
    ('int8', 'bigint'),
    ('NUMERIC', 'numeric'),
    ('NUMERIC(1)', 'numeric(1,0)'),
    ('NUMERIC(12)', 'numeric(12,0)'),
    ('NUMERIC(1000)', 'numeric(1000,0)'),
    ('NUMERIC(12,2)', 'numeric(12,2)'),
    ('NUMERIC(1000,500)', 'numeric(1000,500)'),
    ('DECIMAL', 'numeric'),
    ('dec(9,5)', 'numeric(9,5)'),
    ('REAL', 'real'),
    ('DOUBLE PRECISION', 'double precision'),
    ('FLOAT', 'double precision'),
    ('FLOAT(1)', 'real'),
    ('FLOAT(24)', 'real'),
    ('FLOAT(25)', 'double precision'),
    ('FLOAT(53)', 'double precision'),
    # SERIAL and BIGSERIAL have side effects
    ('MONEY', 'money'),
    ('CHARACTER(1)', 'character(1)'),
    ('CHARACTER VARYING(200000)', 'character varying(200000)'),
    ('CHAR(16)', 'character(16)'),
    ('VARCHAR(256)', 'character varying(256)'),
    ('TEXT', 'text'),
    ('CHAR', 'character(1)'),
    ('CHARACTER VARYING', 'character varying'),
    ('"char"', '"char"'),
    ('name', 'name'),
    ('bytea', 'bytea'),
    ('DATE', 'date'),
    ('TIME', 'time without time zone'),
    ('TIME WITHOUT TIME ZONE', 'time without time zone'),
    ('TIME WITH TIME ZONE', 'time with time zone'),
    ('TIMESTAMP', 'timestamp without time zone'),
    ('TIMESTAMP WITHOUT TIME ZONE', 'timestamp without time zone'),
    ('TIMESTAMP WITH TIME ZONE', 'timestamp with time zone'),
    ('TIME(0)', 'time(0) without time zone'),
    ('TIME(1) WITHOUT TIME ZONE', 'time(1) without time zone'),
    ('TIME(2) WITH TIME ZONE', 'time(2) with time zone'),
    ('TIMESTAMP(3)', 'timestamp(3) without time zone'),
    ('TIMESTAMP(4) WITHOUT TIME ZONE', 'timestamp(4) without time zone'),
    ('TIMESTAMP(5) WITH TIME ZONE', 'timestamp(5) with time zone'),
    ('INTERVAL', 'interval'),
    ('INTERVAL(6)', 'interval(6)'),
    ('INTERVAL YEAR', 'interval year'),
    ('INTERVAL MONTH', 'interval month'),
    ('INTERVAL DAY', 'interval day'),
    ('INTERVAL HOUR', 'interval hour'),
    ('INTERVAL MINUTE', 'interval minute'),
    ('INTERVAL SECOND', 'interval second'),
    ('INTERVAL YEAR TO MONTH', 'interval year to month'),
    ('INTERVAL DAY TO HOUR', 'interval day to hour'),
    ('INTERVAL DAY TO MINUTE', 'interval day to minute'),
    ('INTERVAL DAY TO SECOND', 'interval day to second'),
    ('INTERVAL HOUR TO MINUTE', 'interval hour to minute'),
    ('INTERVAL HOUR TO SECOND', 'interval hour to second'),
    ('INTERVAL MINUTE TO SECOND', 'interval minute to second'),
    ('INTERVAL SECOND(3)', 'interval second(3)'),
    ('INTERVAL HOUR TO SECOND(5)', 'interval hour to second(5)'),
    ('BOOLEAN', 'boolean'),
    ('POINT', 'point'),
    ('LINE', 'line'),
    ('LSEG', 'lseg'),
    ('BOX', 'box'),
    ('PATH', 'path'),
    ('POLYGON', 'polygon'),
    ('CIRCLE', 'circle'),
    ('cidr', 'cidr'),
    ('inet', 'inet'),
    ('macaddr', 'macaddr'),
    ('BIT(2)', 'bit(2)'),
    ('BIT VARYING(100)', 'bit varying(100)'),
    ('BIT', 'bit(1)'),
    ('BIT VARYING', 'bit varying'),
    ('tsvector', 'tsvector'),
    ('tsquery', 'tsquery'),
    ('UUID', 'uuid'),
    ('XML', 'xml')]

DROP_STMT = "DROP TABLE IF EXISTS t1"


class TableToMapTestCase(PyrseasTestCase):
    """Test mapping of created tables"""

    def test_create_table(self):
        "Map a table with two columns"
        ddlstmt = "CREATE TABLE t1 (c1 INTEGER, c2 TEXT)"
        expmap = {'columns': [{'c1': {'type': 'integer'}},
                              {'c2': {'type': 'text'}}]}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], expmap)

    def test_data_types(self):
        "Map a table with columns for (almost) each native PostgreSQL type"
        colstab = []
        colsmap = []
        for colnum, (coltype, maptype) in enumerate(TYPELIST):
            col = "c%d" % (colnum + 1)
            colstab.append("%s %s" % (col, coltype))
            colsmap.append({col: {'type': maptype}})
        ddlstmt = "CREATE TABLE t1 (%s)" % ", ".join(colstab)
        expmap = {'columns': colsmap}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], expmap)

    def test_not_null(self):
        "Map a table with a NOT NULL column"
        ddlstmt = """CREATE TABLE t1 (c1 INTEGER, c2 INTEGER NULL,
                                      c3 INTEGER NOT NULL)"""
        expmap = {'columns': [{'c1': {'type': 'integer'}},
                              {'c2': {'type': 'integer'}},
                              {'c3': {'type': 'integer', 'not_null': True}}]}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], expmap)

    def test_column_defaults(self):
        "Map a table with various types and each with a DEFAULT clause"
        ddlstmt = """CREATE TABLE t1 (c1 INTEGER DEFAULT 12345,
                                      c2 NUMERIC DEFAULT 98.76,
                                      c3 REAL DEFAULT 15e-2,
                                      c4 TEXT DEFAULT 'Abc def',
                                      c5 DATE DEFAULT CURRENT_DATE,
                                      c6 TIMESTAMP WITH TIME ZONE
                                         DEFAULT CURRENT_TIMESTAMP,
                                      c7 BOOLEAN DEFAULT FALSE)"""
        expmap = {'columns': [
                    {'c1': {'type': 'integer', 'default': '12345'}},
                    {'c2': {'type': 'numeric', 'default': '98.76'}},
                    {'c3': {'type': 'real', 'default': '0.15'}},
                    {'c4': {'type': 'text', 'default': "'Abc def'::text"}},
                    {'c5': {'type': 'date', 'default': "('now'::text)::date"}},
                    {'c6': {'type': 'timestamp with time zone',
                            'default': 'now()'}},
                    {'c7': {'type': 'boolean', 'default': 'false'}}]}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], expmap)


class TableToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input schemas"""

    def test_create_table(self):
        "Create a two-column table"
        self.db.execute_commit(DROP_STMT)
        inmap = {'schema public': {'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}]}}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "CREATE TABLE t1 (c1 integer, c2 text)")

    def test_create_table_with_defaults(self):
        "Create a table with two column DEFAULTs, one referring to a SEQUENCE"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("DROP SEQUENCE IF EXISTS t1_c1_seq")
        inmap = {'schema public': {
                'table t1': {
                    'columns': [{'c1': {
                                'type': 'integer',
                                'not_null': True,
                                'default': "nextval('t1_c1_seq'::regclass)"}},
                                {'c2': {'type': 'text', 'not_null': True}},
                                {'c3': {
                                'type': 'date', 'not_null': True,
                                'default': "('now'::text)::date"}}]},
                'sequence t1_c1_seq': {
                    'cache_value': 1, 'increment_by': 1, 'max_value': None,
                    'min_value': None, 'start_value': 1}}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "CREATE TABLE t1 (c1 integer NOT NULL, "
                         "c2 text NOT NULL, "
                         "c3 date NOT NULL DEFAULT ('now'::text)::date)")
        self.assertEqual(fix_indent(dbsql[1]),
                         "CREATE SEQUENCE t1_c1_seq START WITH 1 "
                         "INCREMENT BY 1 NO MAXVALUE NO MINVALUE CACHE 1")
        self.assertEqual(dbsql[2], "ALTER TABLE t1 ALTER COLUMN c1 "
                         "SET DEFAULT nextval('t1_c1_seq'::regclass)")

    def test_bad_table_map(self):
        "Error creating a table with a bad map"
        inmap = {'schema public': {'t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}]}}}
        self.assertRaises(KeyError, self.db.process_map, inmap)

    def test_missing_columns(self):
        "Error creating a table with no columns"
        inmap = {'schema public': {'table t1': {'columns': []}}}
        self.assertRaises(ValueError, self.db.process_map, inmap)

    def test_drop_table(self):
        "Drop an existing table"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE TABLE t1 (c1 INTEGER, c2 TEXT)")
        inmap = {'schema public': {}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, ["DROP TABLE t1"])

    def test_rename_table(self):
        "Rename an existing table"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE TABLE t1 (c1 INTEGER, c2 TEXT)")
        inmap = {'schema public': {'table t2': {
                    'oldname': 't1',
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}]}}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql, ["ALTER TABLE t1 RENAME TO t2"])

    def test_add_columns(self):
        "Add two new columns to a table"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE TABLE t1 (c1 INTEGER, c2 TEXT)")
        inmap = {'schema public': {'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}},
                                {'c3': {'type': 'smallint', 'not_null': True}},
                                {'c4': {'type': 'date',
                                        'default': 'now()'}}]}}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                "ALTER TABLE t1 ADD COLUMN c3 smallint NOT NULL")
        self.assertEqual(fix_indent(dbsql[1]),
                "ALTER TABLE t1 ADD COLUMN c4 date DEFAULT now()")

    def test_set_column_not_null(self):
        "Change a nullable column to NOT NULL"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE TABLE t1 (c1 INTEGER, c2 TEXT)")
        inmap = {'schema public': {'table t1': {
                'columns': [{'c1': {'type': 'integer', 'not_null': True}},
                            {'c2': {'type': 'text'}}]}}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "ALTER TABLE t1 ALTER COLUMN c1 SET NOT NULL")

    def test_change_column_types(self):
        "Change the datatypes of two columns"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE TABLE t1 (c1 INTEGER, c2 TEXT)")
        inmap = {'schema public': {'table t1': {
                    'columns': [{'c1': {'type': 'bigint'}},
                                {'c2': {'type': 'varchar(25)'}}]}}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "ALTER TABLE t1 ALTER COLUMN c1 TYPE bigint")
        self.assertEqual(fix_indent(dbsql[1]),
                         "ALTER TABLE t1 ALTER COLUMN c2 TYPE varchar(25)")

    def test_drop_column(self):
        "Drop a column from a table"
        self.db.execute(DROP_STMT)
        self.db.execute_commit("CREATE TABLE t1 (c1 INTEGER, c2 TEXT, "
                               "c3 SMALLINT, c4 DATE)")
        inmap = {'schema public': {'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}},
                                {'c4': {'type': 'date',
                                        'default': 'now()'}}]}}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]), "ALTER TABLE t1 DROP COLUMN c3")


def suite():
    tests = unittest.TestLoader().loadTestsFromTestCase(TableToMapTestCase)
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            TableToSqlTestCase))
    return tests

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""yamltodb - generate SQL statements to update a PostgreSQL database
to match the schema specified in a YAML file"""

import os
from optparse import OptionParser

import yaml

from pyrseas.dbconn import DbConnection
from pyrseas.database import Database


def main(host='localhost', port=5432):
    """Convert YAML specifications to database DDL."""
    parser = OptionParser("usage: %prog [options] dbname yamlspec")
    parser.add_option('-H', '--host', dest='host',
                      help="database server host or socket directory "
                      "(default %default)")
    parser.add_option('-p', '--port', dest='port', type='int',
                     help="database server port (default %default)")
    parser.add_option('-U', '--username', dest='username',
                     help="database user name (default %default)")
    parser.add_option('-1', '--single-transaction', action='store_true',
                      dest='onetrans',
                      help="wrap commands in BEGIN/COMMIT")

    parser.set_defaults(host=host, port=port, username=os.getenv("USER"))
    (options, args) = parser.parse_args()
    if len(args) > 2:
        parser.error("too many arguments")
    elif len(args) != 2:
        parser.error("missing arguments")
    dbname = args[0]
    yamlspec = args[1]

    db = Database(DbConnection(dbname, options.username, options.host,
                               options.port))
    stmts = db.diff_map(yaml.load(open(yamlspec)))
    if stmts:
        if options.onetrans:
            print "BEGIN;"
        print ";\n".join(stmts) + ';'
        if options.onetrans:
            print "COMMIT;"

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    pyrseas.table
    ~~~~~~~~~~~~~

    This module defines four classes: DbClass derived from
    DbSchemaObject, Sequence and Table derived from DbClass, and
    ClassDict derived from DbObjectDict.
"""
import sys

from pyrseas.dbobject import DbObjectDict, DbSchemaObject
from constraint import CheckConstraint, PrimaryKey, ForeignKey, \
    UniqueConstraint

MAX_BIGINT = 9223372036854775807L


class DbClass(DbSchemaObject):
    """A table, sequence or view"""

    keylist = ['schema', 'name']


class Sequence(DbClass):
    "A sequence generator definition"

    keylist = ['schema', 'name']
    objtype = "SEQUENCE"

    def get_attrs(self, dbconn):
        """Get the attributes for the sequence

        :param dbconn: a DbConnection object
        """
        data = dbconn.fetchone(
            """SELECT start_value, increment_by, max_value, min_value,
                      cache_value
               FROM %s.%s""" % (self.schema, self.name))
        for key, val in data.items():
            setattr(self, key, val)

    def get_owner(self, dbconn):
        """Get the table and column name that owns the sequence

        :param dbconn: a DbConnection object
        """
        data = dbconn.fetchone(
            """SELECT refobjid::regclass, refobjsubid
               FROM pg_depend
               WHERE objid = '%s'::regclass
                 AND refclassid = 'pg_class'::regclass""" % self.qualname())
        if data:
            self.owner_table = tbl = data[0]
            self.owner_column = data[1]
            if '.' in tbl:
                dot = tbl.index('.')
                if self.schema == tbl[:dot]:
                    self.owner_table = tbl[dot + 1:]

    def to_map(self):
        """Convert a sequence definition to a YAML-suitable format

        :return: dictionary
        """
        seq = {}
        for key, val in self.__dict__.items():
            if key in self.keylist:
                continue
            if key == 'max_value' and val == MAX_BIGINT:
                seq[key] = None
            elif key == 'min_value' and val == 1:
                seq[key] = None
            elif val <= sys.maxint:
                seq[key] = int(val)
            else:
                seq[key] = str(val)
        return {self.extern_key(): seq}

    def create(self):
        """Return a SQL statement to CREATE the sequence

        :return: SQL statement
        """
        maxval = self.max_value and ("MAXVALUE %d" % self.max_value) \
            or "NO MAXVALUE"
        minval = self.min_value and ("MINVALUE %d" % self.min_value) \
            or "NO MINVALUE"
        return """CREATE SEQUENCE %s
    START WITH %d
    INCREMENT BY %d
    %s
    %s
    CACHE %d""" % (self.qualname(), self.start_value, self.increment_by,
                   maxval, minval, self.cache_value)

    def add_owner(self):
        """Return statement to ALTER the sequence to indicate its owner table

        :return: SQL statement
        """
        stmts = []
        pth = self.set_search_path()
        if pth:
            stmts.append(pth)
        stmts.append("ALTER SEQUENCE %s OWNED BY %s.%s" % (
                self.name, self.owner_table, self.owner_column))
        return stmts

    def diff_map(self, inseq):
        """Generate SQL to transform an existing sequence

        :param inseq: a YAML map defining the new sequence
        :return: list of SQL statements

        Compares the sequence to an input sequence and generates SQL
        statements to transform it into the one represented by the
        input.
        """
        stmt = ""
        if self.start_value != inseq.start_value:
            stmt += " START WITH %d" % inseq.start_value
        if self.increment_by != inseq.increment_by:
            stmt += " INCREMENT BY %d" % inseq.increment_by
        maxval = self.max_value
        if maxval == MAX_BIGINT:
            maxval = None
        if maxval != inseq.max_value:
            stmt += inseq.max_value and (" MAXVALUE %d" % inseq.max_value) \
                or " NO MAXVALUE"
        minval = self.min_value
        if minval == 1:
            minval = None
        if minval != inseq.min_value:
            stmt += inseq.min_value and (" MINVALUE %d" % inseq.min_value) \
                or " NO MINVALUE"
        if self.cache_value != inseq.cache_value:
            stmt += " CACHE %d" % inseq.cache_value
        if stmt:
            return "ALTER SEQUENCE %s" % self.qualname() + stmt
        return []


class Table(DbClass):
    """A database table definition

    A table is identified by its schema name and table name.  It should
    have a list of columns.  It may have a primary_key, zero or more
    foreign_keys, zero or more unique_constraints, and zero or more
    indexes.
    """

    objtype = "TABLE"

    def column_names(self):
        """Return a list of column names in the table

        :return: list
        """
        return [c.name for c in self.columns]

    def to_map(self, dbschemas):
        """Convert a table to a YAML-suitable format

        :param dbschemas: database dictionary of schemas
        :return: dictionary
        """
        if not hasattr(self, 'columns'):
            return
        cols = []
        for i in range(len(self.columns)):
            cols.append(self.columns[i].to_map())
        tbl = {'columns': cols}
        if hasattr(self, 'description'):
            tbl.update(description=self.description)
        if hasattr(self, 'check_constraints'):
            if not 'check_constraints' in tbl:
                tbl.update(check_constraints={})
            for k in self.check_constraints.values():
                tbl['check_constraints'].update(
                    self.check_constraints[k.name].to_map(self.column_names()))
        if hasattr(self, 'primary_key'):
            tbl.update(primary_key=self.primary_key.to_map(
                    self.column_names()))
        if hasattr(self, 'foreign_keys'):
            if not 'foreign_keys' in tbl:
                tbl['foreign_keys'] = {}
            for k in self.foreign_keys.values():
                tbls = dbschemas[k.ref_schema].tables
                tbl['foreign_keys'].update(self.foreign_keys[k.name].to_map(
                        self.column_names(),
                        tbls[self.foreign_keys[k.name].ref_table]. \
                            column_names()))
        if hasattr(self, 'unique_constraints'):
            if not 'unique_constraints' in tbl:
                tbl.update(unique_constraints={})
            for k in self.unique_constraints.values():
                tbl['unique_constraints'].update(
                    self.unique_constraints[k.name].to_map(
                        self.column_names()))
        if hasattr(self, 'indexes'):
            if not 'indexes' in tbl:
                tbl['indexes'] = {}
            for k in self.indexes.values():
                tbl['indexes'].update(self.indexes[k.name].to_map(
                        self.column_names()))

        return {self.extern_key(): tbl}

    def create(self):
        """Return a SQL statement to CREATE the table

        :return: SQL statement
        """
        cols = []
        for col in self.columns:
            cols.append("    " + col.add())
        return "CREATE TABLE %s (\n%s)" % (self.qualname(), ",\n".join(cols))

    def diff_map(self, intable):
        """Generate SQL to transform an existing table

        :param intable: a YAML map defining the new table
        :return: list of SQL statements

        Compares the table to an input table and generates SQL
        statements to transform it into the one represented by the
        input.
        """
        stmts = []
        if not hasattr(intable, 'columns'):
            raise KeyError("Table '%s' has no columns" % intable.name)
        dbcols = len(self.columns)

        base = "ALTER TABLE %s\n    " % self.qualname()
        # check input columns
        for (num, incol) in enumerate(intable.columns):
            # add new columns
            if num >= dbcols:
                stmts.append(base + "ADD COLUMN %s" % incol.add())
            # check existing columns
            # TODO: more work is needed, for columns out of order
            elif self.columns[num].name == incol.name:
                stmt = self.columns[num].diff_map(incol)
                if stmt:
                    stmts.append(base + stmt)

        return stmts


class ClassDict(DbObjectDict):
    "The collection of tables and similar objects in a database"

    cls = DbClass
    query = \
        """SELECT nspname AS schema, relname AS name, relkind AS kind
           FROM pg_class
                JOIN pg_namespace ON (relnamespace = pg_namespace.oid)
                JOIN pg_roles ON (nspowner = pg_roles.oid)
           WHERE relkind in ('r', 'S')
                 AND (nspname = 'public' OR rolname <> 'postgres')
           ORDER BY nspname, relname"""

    def _from_catalog(self):
        """Initialize the dictionary of tables by querying the catalogs"""
        for table in self.fetch():
            sch, tbl = table.key()
            kind = table.kind
            del table.kind
            if kind == 'r':
                self[(sch, tbl)] = Table(**table.__dict__)
            elif kind == 'S':
                self[(sch, tbl)] = inst = Sequence(**table.__dict__)
                inst.get_attrs(self.dbconn)
                inst.get_owner(self.dbconn)

    def from_map(self, schema, inobjs, newdb):
        """Initalize the dictionary of tables by converting the input map

        :param schema: schema owning the tables
        :param inobjs: YAML map defining the schema objects
        :param newdb: collection of dictionaries defining the database
        """
        for k in inobjs.keys():
            spc = k.find(' ')
            if spc == -1:
                raise KeyError("Unrecognized object type: %s" % k)
            objtype = k[:spc]
            key = k[spc + 1:]
            if objtype == 'table':
                self[(schema.name, key)] = table = Table(
                    schema=schema.name, name=key)
                intable = inobjs[k]
                if not intable:
                    raise ValueError("Table '%s' has no specification" % k)
                try:
                    newdb.columns.from_map(table, intable['columns'])
                except KeyError, exc:
                    exc.args = ("Table '%s' has no columns" % key, )
                    raise
                if 'oldname' in intable:
                    table.oldname = intable['oldname']
                newdb.constraints.from_map(table, intable)
                if 'indexes' in intable:
                    newdb.indexes.from_map(table, intable['indexes'])
            elif objtype == 'sequence':
                self[(schema.name, key)] = seq = Sequence(
                    schema=schema.name, name=key)
                inseq = inobjs[k]
                if not inseq:
                    raise ValueError("Sequence '%s' has no specification" % k)
                for attr, val in inseq.items():
                    setattr(seq, attr, val)
                if 'oldname' in inseq:
                    seq.oldname = inseq['oldname']
            else:
                raise KeyError("Unrecognized object type: %s" % k)

    def link_refs(self, dbcolumns, dbconstrs, dbindexes):
        """Connect columns, constraints, etc. to their respective tables

        :param dbcolumns: dictionary of columns
        :param dbconstrs: dictionary of constraints
        :param dbindexes: dictionary of indexes

        Links each list of table columns in `dbcolumns` to the
        corresponding table. Fills the `foreign_keys`,
        `unique_constraints` and `indexes` dictionaries for each table
        by traversing the `dbconstrs` and `dbindexes` dictionaries,
        which are keyed by schema, table and constraint or index.
        """
        for (sch, tbl) in dbcolumns.keys():
            assert self[(sch, tbl)]
            self[(sch, tbl)].columns = dbcolumns[(sch, tbl)]
        for (sch, tbl) in self.keys():
            if isinstance(self[(sch, tbl)], Sequence):
                seq = self[(sch, tbl)]
                if hasattr(seq, 'owner_table'):
                    if isinstance(seq.owner_column, int):
                        seq.owner_column = self[(sch, seq.owner_table)]. \
                            column_names()[seq.owner_column - 1]
        for (sch, tbl, cns) in dbconstrs.keys():
            constr = dbconstrs[(sch, tbl, cns)]
            if (sch, tbl) not in self:  # check constraints on domains
                continue
            table = self[(sch, tbl)]
            if isinstance(constr, CheckConstraint):
                if not hasattr(table, 'check_constraints'):
                    table.check_constraints = {}
                table.check_constraints.update({cns: constr})
            elif isinstance(constr, PrimaryKey):
                table.primary_key = constr
            elif isinstance(constr, ForeignKey):
                if not hasattr(table, 'foreign_keys'):
                    table.foreign_keys = {}
                # link referenced and referrer
                constr.references = self[(constr.ref_schema, constr.ref_table)]
                # TODO: there can be more than one
                self[(constr.ref_schema, constr.ref_table)].referred_by = \
                    constr
                table.foreign_keys.update({cns: constr})
            elif isinstance(constr, UniqueConstraint):
                if not hasattr(table, 'unique_constraints'):
                    table.unique_constraints = {}
                table.unique_constraints.update({cns: constr})
        for (sch, tbl, idx) in dbindexes.keys():
            assert self[(sch, tbl)]
            table = self[(sch, tbl)]
            if not hasattr(table, 'indexes'):
                table.indexes = {}
            table.indexes.update({idx: dbindexes[(sch, tbl, idx)]})

    def _rename(self, obj, objtype):
        """Process a RENAME"""
        stmt = ''
        oldname = obj.oldname
        try:
            stmt = self[(obj.schema, oldname)].rename(obj.name)
            del self[(obj.schema, oldname)]
        except KeyError, exc:
            exc.args = ("Previous name '%s' for %s '%s' not found" % (
                    oldname, objtype, obj.name), )
            raise
        return stmt

    def diff_map(self, intables):
        """Generate SQL to transform existing tables and sequences

        :param intables: a YAML map defining the new tables/sequences
        :return: list of SQL statements

        Compares the existing table/sequence definitions, as fetched
        from the catalogs, to the input map and generates SQL
        statements to transform the tables/sequences accordingly.
        """
        stmts = []
        # first pass: sequences owned by a table
        for (sch, seq) in intables.keys():
            inseq = intables[(sch, seq)]
            if not isinstance(inseq, Sequence) or \
                    not hasattr(inseq, 'owner_table'):
                continue
            if (sch, seq) not in self:
                if hasattr(inseq, 'oldname'):
                    stmts.append(self._rename(inseq, "sequence"))
                else:
                    # create new sequence
                    stmts.append(inseq.create())

        # check input tables
        for (sch, tbl) in intables.keys():
            intable = intables[(sch, tbl)]
            if not isinstance(intable, Table):
                continue
            # does it exist in the database?
            if (sch, tbl) not in self:
                if hasattr(intable, 'oldname'):
                    stmts.append(self._rename(intable, "table"))
                else:
                    # create new table
                    stmts.append(intable.create())

        # second pass: input sequences not owned by tables
        for (sch, seq) in intables.keys():
            inseq = intables[(sch, seq)]
            if not isinstance(inseq, Sequence):
                continue
            # does it exist in the database?
            if (sch, seq) not in self:
                if hasattr(inseq, 'oldname'):
                    stmts.append(self._rename(inseq, "sequence"))
                elif hasattr(inseq, 'owner_table'):
                    stmts.append(inseq.add_owner())
                else:
                    # create new sequence
                    stmts.append(inseq.create())

        # check database tables and sequences
        for (sch, tbl) in self.keys():
            table = self[(sch, tbl)]
            # if missing, mark it for dropping
            if (sch, tbl) not in intables:
                table.dropped = False
            else:
                # check table/sequence objects
                stmts.append(table.diff_map(intables[(sch, tbl)]))

        # now drop the marked tables
        for (sch, tbl) in self.keys():
            table = self[(sch, tbl)]
            if isinstance(table, Sequence) and hasattr(table, 'owner_table'):
                continue
            if hasattr(table, 'dropped') and not table.dropped:
                # drop subordinate objects first
                if hasattr(table, 'check_constraints'):
                    for chk in table.check_constraints:
                        stmts.append(table.check_constraints[chk].drop())
                if hasattr(table, 'unique_constraints'):
                    for unq in table.unique_constraints:
                        stmts.append(table.unique_constraints[unq].drop())
                if hasattr(table, 'indexes'):
                    for idx in table.indexes:
                        stmts.append(table.indexes[idx].drop())
                if hasattr(table, 'foreign_keys'):
                    for fgn in table.foreign_keys:
                        stmts.append(table.foreign_keys[fgn].drop())
                if hasattr(table, 'primary_key'):
                    # TODO there can be more than one referred_by
                    if hasattr(table, 'referred_by'):
                        stmts.append(table.referred_by.drop())
                    stmts.append(table.primary_key.drop())
                stmts.append(table.drop())

        # last pass to deal with nextval DEFAULTs
        for (sch, tbl) in intables.keys():
            intable = intables[(sch, tbl)]
            if not isinstance(intable, Table):
                continue
            if (sch, tbl) not in self:
                for col in intable.columns:
                    if hasattr(col, 'default') \
                            and col.default.startswith('nextval'):
                        stmts.append(col.set_sequence_default())

        return stmts



# -*- coding: utf-8 -*-
"""
    pyrseas.schema
    ~~~~~~~~~~~~~~

    This defines two classes, Schema and SchemaDict, derived from
    DbObject and DbObjectDict, respectively.
"""
from pyrseas.dbobject import DbObjectDict, DbObject
from table import Table, Sequence

KEY_PREFIX = 'schema '


class Schema(DbObject):
    """A database schema definition, i.e., a named collection of tables,
    views, triggers and other schema objects."""

    keylist = ['name']

    def extern_key(self):
        """Return the key to be used in external maps for this schema

        :return: string
        """
        return KEY_PREFIX + self.name

    def to_map(self, dbschemas):
        """Convert tables, etc., dictionaries to a YAML-suitable format

        :param dbschemas: dictionary of schemas
        :return: dictionary
        """
        key = self.extern_key()
        schema = {key: {}}
        if hasattr(self, 'sequences'):
            seqs = {}
            for seq in self.sequences.keys():
                seqs.update(self.sequences[seq].to_map())
            schema[key].update(seqs)
        if hasattr(self, 'tables'):
            tbls = {}
            for tbl in self.tables.keys():
                tbls.update(self.tables[tbl].to_map(dbschemas))
            schema[key].update(tbls)
        return schema

    def create(self):
        """Return SQL statement to CREATE the schema

        :return: SQL statement
        """
        return "CREATE SCHEMA %s" % self.name

    def drop(self):
        """Return SQL statement to DROP the schema

        :return: SQL statement
        """
        return "DROP SCHEMA %s CASCADE" % self.name

    def rename(self, newname):
        """Return SQL statement to RENAME the schema

        :param newname: the new name for the schema
        :return: SQL statement
        """
        return "ALTER SCHEMA %s RENAME TO %s" % (self.name, newname)


class SchemaDict(DbObjectDict):
    "The collection of schemas in a database.  Minimally, the 'public' schema."

    cls = Schema
    query = \
        """SELECT nspname AS name
           FROM pg_namespace JOIN pg_roles ON (nspowner = pg_roles.oid)
           WHERE nspname = 'public' OR rolname <> 'postgres'
           ORDER BY nspname"""

    def from_map(self, inmap, newdb):
        """Initialize the dictionary of schemas by converting the input map

        :param inmap: the input YAML map defining the schemas
        :param newdb: collection of dictionaries defining the database

        Starts the recursive analysis of the input map and
        construction of the internal collection of dictionaries
        describing the database objects.
        """
        for sch in inmap.keys():
            if not sch.startswith(KEY_PREFIX):
                raise KeyError("Expected named schema, found '%s'" % sch)
            key = sch.split()[1]
            schema = self[key] = Schema(name=key)
            inschema = inmap[sch]
            if inschema:
                if 'oldname' in inschema:
                    schema.oldname = inschema['oldname']
                    del inschema['oldname']
                newdb.tables.from_map(schema, inschema, newdb)

    def link_refs(self, dbtables):
        """Connect tables and sequences to their respective schemas

        :param dbtables: dictionary of tables and sequences

        Fills in the `tables` and `sequences` dictionaries for each
        schema by traversing the `dbtables` dictionary, which is keyed
        by schema and table name.
        """
        for (sch, tbl) in dbtables.keys():
            table = dbtables[(sch, tbl)]
            assert self[sch]
            schema = self[sch]
            if isinstance(table, Table):
                if not hasattr(schema, 'tables'):
                    schema.tables = {}
                schema.tables.update({tbl: table})
            elif isinstance(table, Sequence):
                if not hasattr(schema, 'sequences'):
                    schema.sequences = {}
                schema.sequences.update({tbl: table})

    def to_map(self):
        """Convert the schema dictionary to a regular dictionary

        :return: dictionary

        Invokes the `to_map` method of each schema to construct a
        dictionary of schemas.
        """
        schemas = {}
        for sch in self.keys():
            schemas.update(self[sch].to_map(self))
        return schemas

    def diff_map(self, inschemas):
        """Generate SQL to transform existing schemas

        :param input_map: a YAML map defining the new schemas
        :return: list of SQL statements

        Compares the existing schema definitions, as fetched from the
        catalogs, to the input map and generates SQL statements to
        transform the schemas accordingly.
        """
        stmts = []
        # check input schemas
        for sch in inschemas.keys():
            insch = inschemas[sch]
            # does it exist in the database?
            if sch not in self:
                # check for possible RENAME
                if hasattr(insch, 'oldname'):
                    oldname = insch.oldname
                    try:
                        stmts.append(self[oldname].rename(insch.name))
                        del self[oldname]
                    except KeyError, exc:
                        exc.args = ("Previous name '%s' for schema '%s' "
                                   "not found" % (oldname, insch.name), )
                        raise
                else:
                    # create new schema
                    stmts.append(insch.create())
        # check database schemas
        for sch in self.keys():
            # if missing and not 'public', drop it
            if sch != 'public' and sch not in inschemas:
                stmts.append(self[sch].drop())
        return stmts



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""dbtoyaml - extract the schema of a PostgreSQL database in YAML format"""

import os
from optparse import OptionParser

import yaml

from pyrseas.dbconn import DbConnection
from pyrseas.database import Database


def main(host='localhost', port=5432, schema=None):
    """Convert database table specifications to YAML."""
    parser = OptionParser("usage: %prog [options] dbname")
    parser.add_option('-H', '--host', dest='host',
                      help="database server host or socket directory "
                      "(default %default)")
    parser.add_option('-p', '--port', dest='port', type='int',
                     help="database server port (default %default)")
    parser.add_option('-U', '--username', dest='username',
                     help="database user name (default %default)")
    parser.add_option('-n', '--schema', dest='schema',
                     help="only for named schema (default %default)")
    parser.add_option('-t', '--table', dest='tablist', action='append',
                     help="only for named tables (default all)")

    parser.set_defaults(host=host, port=port, username=os.getenv("USER"),
                        schema=schema)
    (options, args) = parser.parse_args()
    if len(args) > 1:
        parser.error("too many arguments")
    elif len(args) != 1:
        parser.error("database name not specified")
    dbname = args[0]

    db = Database(DbConnection(dbname, options.username, options.host,
                               options.port))
    dbmap = db.to_map()
    # trim the map of schemas/tables not selected
    if options.schema:
        skey = 'schema ' + options.schema
        for sch in dbmap.keys():
            if sch != skey:
                del dbmap[sch]
    if options.tablist:
        ktablist = ['table ' + tbl for tbl in options.tablist]
        for sch in dbmap.keys():
            for tbl in dbmap[sch].keys():
                if tbl not in ktablist:
                    del dbmap[sch][tbl]
            if not dbmap[sch]:
                del dbmap[sch]

    print yaml.dump(dbmap, default_flow_style=False)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    pyrseas.dbobject
    ~~~~~~~~~~~~~~~~

    This defines two low level classes and an intermediate class.
    Most Pyrseas classes are derived from either DbObject or
    DbObjectDict.
"""


class DbObject(object):
    "A single object in a database catalog, e.g., a schema, a table, a column"

    def __init__(self, **attrs):
        """Initialize the catalog object from a dictionary of attributes

        :param attrs: the dictionary of attributes

        Attributes without a value are discarded.
        """
        for key, val in attrs.items():
            if val:
                setattr(self, key, val)

    def key(self):
        """Return a tuple that identifies the database object

        :return: a single value or a tuple
        """
        lst = [getattr(self, k) for k in self.keylist]
        return len(lst) == 1 and lst[0] or tuple(lst)


class DbSchemaObject(DbObject):
    "A database object that is owned by a certain schema"

    def extern_key(self):
        """Return the key to be used in external maps for this object

        :return: string
        """
        return '%s %s' % (self.objtype.lower(), self.name)

    def qualname(self):
        """Return the schema-qualified name of the object

        :return: string

        No qualification is used if the schema is 'public'.
        """
        return self.schema == 'public' and self.name \
            or "%s.%s" % (self.schema, self.name)

    def unqualify(self):
        """Adjust the schema and table name if the latter is qualified"""
        if hasattr(self, 'table') and '.' in self.table:
            tbl = self.table
            dot = tbl.index('.')
            if self.schema == tbl[:dot]:
                self.table = tbl[dot + 1:]

    def drop(self):
        """Return a SQL DROP statement for the object

        :return: SQL statement
        """
        if not hasattr(self, 'dropped') or not self.dropped:
            self.dropped = True
            return "DROP %s %s" % (self.objtype, self.qualname())
        return []

    def rename(self, newname):
        """Return a SQL ALTER statement to RENAME the object

        :return: SQL statement
        """
        return "ALTER %s %s RENAME TO %s" % (self.objtype, self.qualname(),
                                             newname)

    def set_search_path(self):
        """Return a SQL SET search_path if not in the 'public' schema"""
        stmt = ''
        if self.schema != 'public':
            stmt = "SET search_path TO %s, pg_catalog" % self.schema
        return stmt


class DbObjectDict(dict):
    """A dictionary of database objects, all of the same type"""

    cls = DbObject
    query = ''

    def __init__(self, dbconn=None):
        """Initialize the dictionary

        :param dbconn: a DbConnection object

        If dbconn is not None, the _from_catalog method is called to
        initialize the dictionary from the catalogs.
        """
        dict.__init__(self)
        self.dbconn = dbconn
        if dbconn:
            self._from_catalog()

    def _from_catalog(self):
        """Initialize the dictionary by querying the catalogs

        This is may be overriden by derived classes as needed.
        """
        for obj in self.fetch():
            self[obj.key()] = obj

    def fetch(self):
        """Fetch all objects from the catalogs using the class query

        :return: list of self.cls objects
        """
        if not self.dbconn.conn:
            self.dbconn.connect()
        data = self.dbconn.fetchall(self.query)
        return [self.cls(**dict(row)) for row in data]



# -*- coding: utf-8 -*-
"""Utility functions and classes for testing Pyrseas"""

import os
from unittest import TestCase

from psycopg2 import connect
from psycopg2.extras import DictConnection
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT

from pyrseas.dbconn import DbConnection
from pyrseas.database import Database


def fix_indent(stmt):
    "Fix specifications which are in a new line with indentation"
    return stmt.replace('\n    ', ' ').replace('( ', '(')


def pgconnect(dbname, user=None, host='localhost', port=5432):
    "Connect to a Postgres database using psycopg2"
    return connect("host=%s port=%d dbname=%s user=%s" % (
            host, port, dbname, user or os.getenv("USER")),
                     connection_factory=DictConnection)


def pgexecute(dbconn, query):
    "Execute a query using a cursor"
    curs = dbconn.cursor()
    try:
        curs.execute(query)
    except:
        curs.close()
        dbconn.rollback()
        raise
    return curs


def pgexecute_auto(dbconn, query):
    "Execute a query using a cursor with autocommit enabled"
    isolation_level = dbconn.isolation_level
    dbconn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
    curs = pgexecute(dbconn, query)
    dbconn.set_isolation_level(isolation_level)
    return curs

DBNAME = 'pyrseas_testdb'

CREATE_DDL = "CREATE DATABASE %s TEMPLATE = template0"


class PostgresDb(object):
    """A PostgreSQL database connection

    This is separate from the one used by DbConnection, because the
    tests need to create and drop databases and other objects,
    independently.
    """
    def __init__(self, name):
        self.name = name
        self.conn = None

    def connect(self):
        """Connect to the database.  If we're not already connected we
        first connect to 'postgres' and see if the given database exists.
        If it doesn't, we create and then connect to it."""
        if not self.conn:
            conn = pgconnect('postgres')
            curs = pgexecute(conn,
                             "SELECT 1 FROM pg_database WHERE datname = '%s'" %
                             self.name)
            row = curs.fetchone()
            if not row:
                curs.close()
                curs = pgexecute_auto(conn, CREATE_DDL % self.name)
                curs.close()
                conn.close()
            self.conn = pgconnect(self.name)

    def close(self):
        "Close the connection if still open"
        if not self.conn:
            return ValueError
        self.conn.close()

    def create(self):
        "Drop the database if it exists and re-create it"
        conn = pgconnect('postgres')
        curs = pgexecute_auto(conn, "DROP DATABASE IF EXISTS %s" % self.name)
        curs = pgexecute_auto(conn, CREATE_DDL % self.name)
        curs.close()
        conn.close()

    def clear(self):
        "Drop tables and other objects"
        query = \
            """SELECT nspname, relname, relkind FROM pg_class
                      JOIN pg_namespace ON (relnamespace = pg_namespace.oid)
                      JOIN pg_roles ON (nspowner = pg_roles.oid)
               WHERE relkind in ('r', 'S')
                     AND (nspname = 'public' OR rolname <> 'postgres')"""
        curs = pgexecute(self.conn, query)
        objs = curs.fetchall()
        curs.close()
        self.conn.rollback()
        for obj in objs:
            if obj['relkind'] == 'r':
                self.execute("DROP TABLE %s.%s CASCADE" % (obj[0], obj[1]))
            elif obj['relkind'] == 'S':
                self.execute("DROP SEQUENCE %s.%s CASCADE" % (obj[0], obj[1]))
        self.conn.commit()

    def drop(self):
        "Drop the database"
        conn = pgconnect('postgres')
        curs = pgexecute_auto(conn, "DROP DATABASE %s" % self.name)
        curs.close()
        conn.close()

    def execute(self, stmt):
        "Execute a DDL statement"
        #print "execute:", stmt
        curs = pgexecute(self.conn, stmt)
        curs.close()

    def execute_commit(self, stmt):
        "Execute a DDL statement and commit"
        self.execute(stmt)
        self.conn.commit()

    def execute_and_map(self, ddlstmt):
        "Execute a DDL statement, commit, and return a map of the database"
        self.execute(ddlstmt)
        self.conn.commit()
        db = Database(DbConnection(self.name))
        return db.to_map()

    def process_map(self, input_map):
        """Process an input map and return the SQL statements necessary to
        convert the database to match the map."""
        db = Database(DbConnection(self.name))
        stmts = db.diff_map(input_map)
        return stmts


class PyrseasTestCase(TestCase):
    """Base class for most test cases"""

    def setUp(self):
        self.db = PostgresDb(DBNAME)
        self.db.connect()
        self.db.clear()

    def tearDown(self):
        self.db.close()



# -*- coding: utf-8 -*-
"""
    pyrseas.dbconn
    ~~~~~~~~~~~~~~

    A `DbConnection` is a helper class representing a connection to a
    PostgreSQL database.
"""

import os

from psycopg2 import connect
from psycopg2.extras import DictConnection


class DbConnection(object):
    """A database connection, possibly disconnected"""

    def __init__(self, dbname, user=None, host='localhost', port=5432):
        """Initialize the connection information

        :param dbname: database name
        :param user: user name
        :param host: host name
        :param port: host port number
        """
        self.dbname = dbname
        self.user = user
        self.host = host
        self.port = port
        self.conn = None

    def connect(self):
        """Connect to the database

        If user is None, the USER environment variable is used
        instead. The password is either not required or supplied by
        other means, e.g., a $HOME/.pgpass file.
        """
        self.conn = connect("host=%s port=%d dbname=%s user=%s" % (
                self.host, self.port, self.dbname,
                self.user or os.getenv("USER")),
                            connection_factory=DictConnection)
        self._execute("set search_path to public, pg_catalog")

    def _execute(self, query):
        """Create a cursor, execute a query and return the cursor"""
        curs = self.conn.cursor()
        try:
            curs.execute(query)
        except Exception, exc:
            exc.args += (query, )
            raise
        return curs

    def fetchone(self, query):
        """Execute a single row SELECT query and return data

        :param query: a SELECT query to be executed
        :return: a psycopg2 DictRow

        The cursor is closed and a rollback is issued.
        """
        curs = self._execute(query)
        data = curs.fetchone()
        curs.close()
        self.conn.rollback()
        return data

    def fetchall(self, query):
        """Execute a SELECT query and return data

        :param query: a SELECT query to be executed
        :return: a list of psycopg2 DictRow's

        The cursor is closed and a rollback is issued.
        """
        curs = self._execute(query)
        data = curs.fetchall()
        curs.close()
        self.conn.rollback()
        return data



# -*- coding: utf-8 -*-
"""Test indexes"""

import unittest

from utils import PyrseasTestCase, fix_indent


class IndexToMapTestCase(PyrseasTestCase):
    """Test mapping of created indexes"""

    def test_index_1(self):
        "Map a single-column index"
        self.db.execute("CREATE TABLE t1 (c1 INTEGER, c2 TEXT)")
        ddlstmt = "CREATE INDEX t1_idx ON t1 (c1)"
        expmap = {'columns': [{'c1': {'type': 'integer'}},
                              {'c2': {'type': 'text'}}],
                  'indexes': {'t1_idx': {'columns': ['c1'],
                                          'access_method': 'btree'}}}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], expmap)

    def test_index_2(self):
        "Map a two-column index"
        self.db.execute("CREATE TABLE t1 (c1 INTEGER, c2 CHAR(5), c3 TEXT)")
        ddlstmt = "CREATE UNIQUE INDEX t1_idx ON t1 (c1, c2)"
        expmap = {'columns': [{'c1': {'type': 'integer'}},
                              {'c2': {'type': 'character(5)'}},
                              {'c3': {'type': 'text'}}],
                  'indexes': {'t1_idx': {'columns': ['c1', 'c2'],
                                          'unique': True,
                                          'access_method': 'btree'}}}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], expmap)

    def test_index_3(self):
        "Map a table with a unique index and a non-unique GIN index"
        self.db.execute("CREATE TABLE t1 (c1 INTEGER, c2 CHAR(5), "
                        "c3 tsvector)")
        self.db.execute("CREATE UNIQUE INDEX t1_idx_1 ON t1 (c1, c2)")
        ddlstmt = "CREATE INDEX t1_idx_2 ON t1 USING gin (c3)"
        expmap = {'columns': [{'c1': {'type': 'integer'}},
                              {'c2': {'type': 'character(5)'}},
                              {'c3': {'type': 'tsvector'}}],
                  'indexes': {'t1_idx_1': {'columns': ['c1', 'c2'],
                                            'unique': True,
                                            'access_method': 'btree'},
                              't1_idx_2': {'columns': ['c3'],
                                            'access_method': 'gin'}}}
        dbmap = self.db.execute_and_map(ddlstmt)
        self.assertEqual(dbmap['schema public']['table t1'], expmap)


class IndexToSqlTestCase(PyrseasTestCase):
    """Test SQL generation from input indexes"""

    def test_create_table_with_index(self):
        "Create new table with a single column index"
        self.db.execute_commit("DROP TABLE IF EXISTS t1")
        inmap = {'schema public': {'table t1': {
                    'columns': [{'c1': {'type': 'integer'}},
                                {'c2': {'type': 'text'}}],
                    'indexes': {'t1_idx': {
                            'columns': ['c1'],
                            'access_method': 'btree'}}}}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(fix_indent(dbsql[0]),
                         "CREATE TABLE t1 (c1 integer, c2 text)")
        self.assertEqual(dbsql[1],
                         "CREATE INDEX t1_idx ON t1 USING btree (c1)")

    def test_add_index(self):
        "Add a two-column unique index to an existing table"
        self.db.execute("DROP TABLE IF EXISTS t1")
        self.db.execute_commit("CREATE TABLE t1 (c1 INTEGER NOT NULL, "
                        "c2 INTEGER NOT NULL, c3 TEXT)")
        inmap = {'schema public': {'table t1': {
                    'columns': [
                        {'c1': {'type': 'integer', 'not_null': True}},
                        {'c2': {'type': 'integer', 'not_null': True}},
                        {'c3': {'type': 'text'}}],
                    'indexes': {'t1_idx': {
                            'columns': ['c2', 'c1'],
                            'unique': True}}}}}
        dbsql = self.db.process_map(inmap)
        self.assertEqual(dbsql,
                         ["CREATE UNIQUE INDEX t1_idx ON t1 (c2, c1)"])


def suite():
    tests = unittest.TestLoader().loadTestsFromTestCase(IndexToMapTestCase)
    tests.addTest(unittest.TestLoader().loadTestsFromTestCase(
            IndexToSqlTestCase))
    return tests

if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""
    pyrseas.constraint
    ~~~~~~~~~~~~~~~~~~

    This module defines six classes: Constraint derived from
    DbSchemaObject, CheckConstraint, PrimaryKey, ForeignKey and
    UniqueConstraint derived from Constraint, and ConstraintDict
    derived from DbObjectDict.
"""
from pyrseas.dbobject import DbObjectDict, DbSchemaObject


class Constraint(DbSchemaObject):
    """A constraint definition, such as a primary key, foreign key or
       unique constraint"""

    keylist = ['schema', 'table', 'name']

    def key_columns(self):
        """Return comma-separated list of key column names

        :return: string
        """
        return ", ".join(self.keycols)

    def _qualtable(self):
        """Return a schema-qualified name for a newly constructed object"""
        return DbSchemaObject(schema=self.schema, name=self.table).qualname()

    def add(self):
        """Return string to add the constraint via ALTER TABLE

        :return: SQL statement

        Works as is for primary keys and unique constraints but has
        to be overridden for check constraints and foreign keys.
        """
        return "ALTER TABLE %s ADD CONSTRAINT %s %s (%s)" % (
            DbSchemaObject(schema=self.schema, name=self.table).qualname(),
            self.name,
            self.objtype, self.key_columns())

    def drop(self):
        """Return string to drop the constraint via ALTER TABLE

        :return: SQL statement
        """
        if not hasattr(self, 'dropped') or not self.dropped:
            self.dropped = True
            return "ALTER TABLE %s DROP CONSTRAINT %s" % (
                self._qualtable(), self.qualname())
        return []


class CheckConstraint(Constraint):
    "A check constraint definition"

    objtype = "CHECK"

    def to_map(self, dbcols):
        """Convert a check constraint definition to a YAML-suitable format

        :param dbcols: dictionary of dbobject columns
        :return: dictionary
        """
        dct = self.__dict__.copy()
        for k in self.keylist:
            del dct[k]
        dct['columns'] = [dbcols[k - 1] for k in self.keycols]
        del dct['keycols']
        return {self.name: dct}

    def add(self):
        """Return string to add the CHECK constraint via ALTER TABLE

        :return: SQL statement
        """
        return "ALTER TABLE %s ADD CONSTRAINT %s %s (%s)" % (
            self._qualtable(), self.name, self.objtype, self.expression)

    def diff_map(self, inchk):
        """Generate SQL to transform an existing CHECK constraint

        :param inchk: a YAML map defining the new CHECK constraint
        :return: list of SQL statements

        Compares the CHECK constraint to an input constraint and generates
        SQL statements to transform it into the one represented by the
        input.
        """
        stmts = []
        # TODO: to be implemented
        return stmts


class PrimaryKey(Constraint):
    "A primary key constraint definition"

    objtype = "PRIMARY KEY"

    def to_map(self, dbcols):
        """Convert a primary key definition to a YAML-suitable format

        :param dbcols: dictionary of dbobject columns
        :return: dictionary
        """
        dct = self.__dict__.copy()
        for k in self.keylist:
            del dct[k]
        dct['columns'] = [dbcols[k - 1] for k in self.keycols]
        del dct['keycols']
        return {self.name: dct}

    def diff_map(self, inpk):
        """Generate SQL to transform an existing primary key

        :param inpk: a YAML map defining the new primary key
        :return: list of SQL statements

        Compares the primary key to an input primary key and generates
        SQL statements to transform it into the one represented by the
        input.
        """
        stmts = []
        # TODO: to be implemented (via ALTER DROP and ALTER ADD)
        return stmts


class ForeignKey(Constraint):
    "A foreign key constraint definition"

    objtype = "FOREIGN KEY"

    def ref_columns(self):
        """Return comma-separated list of reference column names

        :return: string
        """
        return ", ".join(self.ref_cols)

    def to_map(self, dbcols, refcols):
        """Convert a foreign key definition to a YAML-suitable format

        :param dbcols: dictionary of dbobject columns
        :return: dictionary
        """
        dct = self.__dict__.copy()
        for k in self.keylist:
            del dct[k]
        dct['columns'] = [dbcols[k - 1] for k in self.keycols]
        del dct['keycols']
        refsch = hasattr(self, 'ref_schema') and self.ref_schema or self.schema
        ref_cols = [refcols[k - 1] for k in self.ref_cols]
        dct['references'] = {'table': dct['ref_table'], 'columns': ref_cols}
        if 'ref_schema' in dct:
            dct['references'].update(schema=dct['ref_schema'])
            del dct['ref_schema']
        del dct['ref_table'], dct['ref_cols']
        return {self.name: dct}

    def add(self):
        """Return string to add the foreign key via ALTER TABLE

        :return: SQL statement
        """
        return "ALTER TABLE %s ADD CONSTRAINT %s FOREIGN KEY (%s) " \
            "REFERENCES %s (%s)" % (
            self._qualtable(), self.name, self.key_columns(),
            self.references.qualname(), self.ref_columns())

    def diff_map(self, infk):
        """Generate SQL to transform an existing foreign key

        :param infk: a YAML map defining the new foreign key
        :return: list of SQL statements

        Compares the foreign key to an input foreign key and generates
        SQL statements to transform it into the one represented by the
        input.
        """
        stmts = []
        # TODO: to be implemented (via ALTER DROP and ALTER ADD)
        return stmts


class UniqueConstraint(Constraint):
    "A unique constraint definition"

    objtype = "UNIQUE"

    def to_map(self, dbcols):
        """Convert a unique constraint definition to a YAML-suitable format

        :param dbcols: dictionary of dbobject columns
        :return: dictionary
        """
        dct = self.__dict__.copy()
        for k in self.keylist:
            del dct[k]

        dct['columns'] = []
        dct['columns'] = [dbcols[k - 1] for k in self.keycols]
        del dct['keycols']
        return {self.name: dct}

    def diff_map(self, inuc):
        """Generate SQL to transform an existing unique constraint

        :param inuc: a YAML map defining the new unique constraint
        :return: list of SQL statements

        Compares the unique constraint to an input unique constraint
        and generates SQL statements to transform it into the one
        represented by the input.
        """
        stmts = []
        # TODO: to be implemented (via ALTER DROP and ALTER ADD)
        return stmts


class ConstraintDict(DbObjectDict):
    "The collection of table or column constraints in a database"

    cls = Constraint
    query = \
        """SELECT nspname AS schema, conrelid::regclass AS table,
                  conname AS name, contype AS type, conkey AS keycols,
                  confrelid::regclass AS ref_table, confkey AS ref_cols,
                  consrc AS expression, amname AS access_method
           FROM pg_constraint
                JOIN pg_namespace ON (connamespace = pg_namespace.oid)
                JOIN pg_roles ON (nspowner = pg_roles.oid)
                LEFT JOIN pg_class on (conname = relname)
                LEFT JOIN pg_am on (relam = pg_am.oid)
           WHERE (nspname = 'public' OR rolname <> 'postgres')
           ORDER BY schema, 2, name"""

    def _from_catalog(self):
        """Initialize the dictionary of constraints by querying the catalogs"""
        for constr in self.fetch():
            constr.unqualify()
            sch, tbl, cns = constr.key()
            constr_type = constr.type
            del constr.type
            if constr_type == 'c':
                del constr.ref_table
                self[(sch, tbl, cns)] = CheckConstraint(**constr.__dict__)
            elif constr_type == 'p':
                del constr.ref_table
                self[(sch, tbl, cns)] = PrimaryKey(**constr.__dict__)
            elif constr_type == 'f':
                # normalize reference schema/table
                reftbl = constr.ref_table
                if '.' in reftbl:
                    dot = reftbl.index('.')
                    constr.ref_table = reftbl[dot + 1:]
                    constr.ref_schema = reftbl[:dot]
                else:
                    constr.ref_schema = constr.schema
                self[(sch, tbl, cns)] = ForeignKey(**constr.__dict__)
            elif constr_type == 'u':
                del constr.ref_table
                self[(sch, tbl, cns)] = UniqueConstraint(**constr.__dict__)

    def from_map(self, table, inconstrs):
        """Initialize the dictionary of constraints by converting the input map

        :param table: table affected by the constraints
        :param inconstrs: YAML map defining the constraints
        """
        if 'check_constraints' in inconstrs:
            chks = inconstrs['check_constraints']
            for cns in chks.keys():
                check = CheckConstraint(table=table.name, schema=table.schema,
                                      name=cns)
                val = chks[cns]
                try:
                    check.expression = val['expression']
                except KeyError, exc:
                    exc.args = ("Constraint '%s' is missing expression"
                                % cns, )
                    raise
                if check.expression[0] == '(' and check.expression[-1] == ')':
                    check.expression = check.expression[1:-1]
                if 'columns' in val:
                    check.keycols = val['columns']
                self[(table.schema, table.name, cns)] = check
        if 'primary_key' in inconstrs:
            cns = inconstrs['primary_key'].keys()[0]
            pkey = PrimaryKey(table=table.name, schema=table.schema,
                              name=cns)
            val = inconstrs['primary_key'][cns]
            try:
                pkey.keycols = val['columns']
            except KeyError, exc:
                exc.args = ("Constraint '%s' is missing columns" % cns, )
                raise
            if 'access_method' in val:
                pkey.access_method = val['access_method']
            self[(table.schema, table.name, cns)] = pkey
        if 'foreign_keys' in inconstrs:
            fkeys = inconstrs['foreign_keys']
            for cns in fkeys.keys():
                fkey = ForeignKey(table=table.name, schema=table.schema,
                                      name=cns)
                val = fkeys[cns]
                try:
                    fkey.keycols = val['columns']
                except KeyError, exc:
                    exc.args = ("Constraint '%s' is missing columns" % cns, )
                    raise
                try:
                    refs = val['references']
                except KeyError, exc:
                    exc.args = ("Constraint '%s' missing references" % cns, )
                    raise
                try:
                    fkey.ref_table = refs['table']
                except KeyError, exc:
                    exc.args = ("Constraint '%s' missing table reference"
                                % cns, )
                    raise
                try:
                    fkey.ref_cols = refs['columns']
                except KeyError, exc:
                    exc.args = ("Constraint '%s' missing reference columns"
                                    % cns, )
                    raise
                sch = table.schema
                if 'schema' in refs:
                    sch = refs['schema']
                fkey.ref_schema = sch
                self[(table.schema, table.name, cns)] = fkey
        if 'unique_constraints' in inconstrs:
            uconstrs = inconstrs['unique_constraints']
            for cns in uconstrs.keys():
                unq = UniqueConstraint(table=table.name, schema=table.schema,
                                      name=cns)
                val = uconstrs[cns]
                try:
                    unq.keycols = val['columns']
                except KeyError, exc:
                    exc.args = ("Constraint '%s' is missing columns" % cns, )
                    raise
                if 'access_method' in val:
                    unq.access_method = val['access_method']
                self[(table.schema, table.name, cns)] = unq

    def diff_map(self, inconstrs):
        """Generate SQL to transform existing constraints

        :param inconstrs: a YAML map defining the new constraints
        :return: list of SQL statements

        Compares the existing constraint definitions, as fetched from
        the catalogs, to the input map and generates SQL statements to
        transform the constraints accordingly.
        """
        stmts = []
        # foreign keys are processed in a second pass
        # constraints cannot be renamed
        for turn in (1, 2):
            # check database constraints
            for (sch, tbl, cns) in self.keys():
                constr = self[(sch, tbl, cns)]
                if isinstance(constr, ForeignKey):
                    if turn == 1:
                        continue
                elif turn == 2:
                    continue
                # if missing, drop it
                if (sch, tbl, cns) not in inconstrs:
                    stmts.append(constr.drop())
            # check input constraints
            for (sch, tbl, cns) in inconstrs.keys():
                inconstr = inconstrs[(sch, tbl, cns)]
                if isinstance(inconstr, ForeignKey):
                    if turn == 1:
                        continue
                elif turn == 2:
                    continue
                # does it exist in the database?
                if (sch, tbl, cns) not in self:
                    # add the new constraint
                    stmts.append(inconstr.add())
                else:
                    # check constraint objects
                    stmts.append(self[(sch, tbl, cns)].diff_map(inconstr))

        return stmts
